/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2010 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.android.controller;

import java.util.Enumeration;
import java.util.Vector;

import android.os.Bundle;
import android.content.Context;
import android.content.ContentResolver;
import android.accounts.Account;
import android.os.Handler;

import com.funambol.android.AndroidAppSyncSource;
import com.funambol.android.activities.AndroidDisplayManager;
import com.funambol.android.activities.AndroidHomeScreen;

import com.funambol.platform.NetworkStatus;
import com.funambol.client.controller.HomeScreenController;
import com.funambol.client.controller.Controller;
import com.funambol.client.source.AppSyncSource;
import com.funambol.client.ui.HomeScreen;
import com.funambol.client.ui.DisplayManager;
import com.funambol.syncml.spds.SyncSource;
import com.funambol.syncml.spds.SyncListener;
import com.funambol.syncml.spds.SyncReport;
import com.funambol.syncml.spds.SyncException;
import com.funambol.util.Log;

public class AndroidHomeScreenController extends HomeScreenController {

    private static final String TAG = "AndroidHomeScreenController";

    public static final String SYNC_TYPE = "SyncType";
    public static final String REFRESH   = "Refresh";
    public static final String REFRESH_DIRECTION = "RefreshDirection";
    public static final String AUTHORITY_TYPE = "AuthorityType";

    private Context context;

    private boolean syncAll = false;
    private String  syncAllType = MANUAL;
    private Vector  syncAllSources = null;
    private boolean refresh        = false;
    private int     refreshDirection;

    private NetworkStatus networkStatus;

    private AutoSyncSwitcher autoSyncSwitcher;

    public AndroidHomeScreenController(Context context, Controller controller, HomeScreen homeScreen) {
        super(controller, homeScreen);
        this.context = context;
        engine.setSpawnThread(false);
        networkStatus = new NetworkStatus(context);
        engine.setNetworkStatus(networkStatus);
        autoSyncSwitcher = new AutoSyncSwitcher(context, new Handler(), this);
    }

    /**
     * This method allows a sync to be cancelled. Cancelling a sync on Android
     * is a rather tricky operation, because there are several entry points and
     * different status in which the sync can be.
     * The following cases need to be addressed:
     *
     * <ul>
     *   <li> the sync is pending (meaning it was requested to the native sync
     *        subsystem, but it has been queued) </li>
     *
     *   <li> the sync is active (the native sync subsystem triggered it) but the
     *        SyncManager has not been invoked yet </li>
     *
     *   <li> the sync is active and the SyncManager has been invoked </li>
     *
     *   <li> the sync is cancelled via the native application </li>
     * </ul>
     *
     * For each case, we must make sure that the following things happen:
     *
     * <ul>
     *   <li> The home screen is properly refreshed </li>
     *   <li> The SyncEngine is unloked and the sync terminated </li>
     * </ul>
     */
    @Override
    public void cancelSync() {
        Log.trace(TAG, "Cancelling sync " + isSynchronizing() + " currentSource=" + currentSource);
        setCancel(true);

        // If a sync is active, we cancel it
        Account account = AndroidController.getNativeAccount();
        
        AndroidAppSyncSource appSource = isAnySourceSyncing();
        boolean cancelled = false;
        if (appSource != null) {
            String authority = appSource.getAuthority();
            ContentResolver.cancelSync(account, authority);
            cancelled = true;
        }

        // Remove all the sources in the pending state
        Enumeration sources = appSyncSourceManager.getWorkingSources();
        while(sources.hasMoreElements()) {
            appSource = (AndroidAppSyncSource)sources.nextElement();
            String authority = appSource.getAuthority();
            if (authority != null) {
                if (ContentResolver.isSyncPending(account, authority)) {
                    ContentResolver.cancelSync(account, authority);
                    if (!cancelled) {
                        forceSyncCancel(appSource);
                        cancelled = true;
                    }
                }
            }
        }

        if (!cancelled) {
            super.cancelSync();
        }
    }

    public boolean isSynchronizing() {
        if (isAnySourcePending() != null) {
            return true;
        } else {
            return super.isSynchronizing();
        }
    }

    public synchronized void synchronize(String syncType, Vector syncSources,
                                         int delay, boolean fromOutside) {

        Log.info(TAG, "synchronize " + syncType);

        if (!PUSH.equals(syncType) && isSynchronizing()) {
            Log.info(TAG, "A sync is already in progress");
            if(MANUAL.equals(syncType)) {
                showSyncInProgressMessage();
            }
            return;
        }
        forceSynchronization(syncType, syncSources, false, 0, delay, fromOutside);
    }

    public void showConfigurationScreen() {
        if (syncAll) {
            showSyncInProgressMessage();
        } else {
            super.showConfigurationScreen();
        }
    }

    /**
     * This method allows a sync to be cancelled. Cancelling a sync on Android
     * is a rather tricky operation, because there are several entry points and
     * different status in which the sync can be.
     * The following cases need to be addressed:
     *
     * <ul>
     *   <li> the sync is pending (meaning it was requested to the native sync
     *        subsystem, but it has been queued) </li>
     *
     *   <li> the sync is active (the native sync subsystem triggered it) but the
     *        SyncManager has not been invoked yet </li>
     *
     *   <li> the sync is active and the SyncManager has been invoked </li>
     *
     *   <li> the sync is cancelled via the native application </li>
     * </ul>
     *
     * For each case, we must make sure that the following things happen:
     *
     * <ul>
     *   <li> The home screen is properly refreshed </li>
     *   <li> The AndroidSyncEngine is unloked and the sync terminated </li>
     * </ul>
     */
    public void forceSyncCancel(AppSyncSource appSource) {
        if (engine.isSynchronizing()) {
            super.cancelSync();
        } else {
            // SyncManager won't be able to throw any exception and we
            // must simulate a sync termination here
            SyncSource source = appSource.getSyncSource();
            SyncListener listener = source.getListener();
            SyncReport syncReport = new SyncReport(source);
            //syncReport.setRequestedSyncMode(syncMode);
            syncReport.setLocUri(source.getName());
            syncReport.setRemoteUri(source.getSourceUri());
            syncReport.setSyncStatus(SyncListener.CANCELLED);
            listener.endSession(syncReport);
            changeSyncLabelsOnSyncEnded();
            unlockHomeScreen();
        }
    }

    public void syncSingleSource(String syncType, AppSyncSource appSource) {
        syncAll = false;
        syncAllSources = null;

        syncSource(syncType, appSource);
    }

    /**
     * This method is invoked when a sync is triggered from within the
     * application. This means the user clicked on a sync button (either a
     * source or sync all).
     * If the source is configured to use the sync adapter, then a native sync
     * is triggered, otherwise the SynchronizationController gets invoked
     * directly.
     */
    protected void syncSource(String syncType, AppSyncSource appSource) {

        if (networkStatus != null && !networkStatus.isConnected()) {
            if (networkStatus.isRadioOff()) {
                noConnection();
            } else {
                noSignal();
            }
            //The sync all button could have been pressed when a running sync
            //wasn't detected. This avoid the user not allowed to enter the
            //settings when the network is down. 
            syncAll = false;
            return;
        }

        // A sync was triggered from our UI, fire a sync via the native sync
        Account account = AndroidController.getNativeAccount();

        Log.debug(TAG, "syncSource " + appSource.getName());
        AndroidAppSyncSource aAppSource = (AndroidAppSyncSource)appSource;
        if (aAppSource.getSyncMethod() == AndroidAppSyncSource.SYNC_ADAPTER) {
            Bundle b = new Bundle();
            b.putBoolean(ContentResolver.SYNC_EXTRAS_FORCE, true);
            b.putString(SYNC_TYPE, syncType);
            if(aAppSource.getAuthorityType() != null) {
                b.putString(AUTHORITY_TYPE, aAppSource.getAuthorityType());
            }
            String authority = aAppSource.getAuthority();
            if (authority != null) {
                ContentResolver.requestSync(account, authority, b);
            } else {
                Log.error(TAG, "Cannot sync source with unknown authority");
            }
        } else {
            super.syncSource(syncType, aAppSource);
        }
    }

    public void refresh(int mask, int direction) {
        // A sync was triggered from our UI, fire a sync via the native sync
        Enumeration sources = appSyncSourceManager.getEnabledAndWorkingSources();
        Vector appSources = new Vector();
        while(sources.hasMoreElements()) {
            AndroidAppSyncSource appSource = (AndroidAppSyncSource)sources.nextElement();
            if ((appSource.getId() & mask) != 0) {
                Log.debug(TAG, "refreshSource " + appSource.getName());
                appSources.addElement(appSource);
            }
        }

        if (appSources.size() > 0) {

            AndroidAppSyncSource appSource = (AndroidAppSyncSource)appSources.elementAt(0);
            appSources.removeElementAt(0);

            syncAllSources = appSources;
            syncAll = true;
            refresh = true;
            refreshDirection = direction;

            Account account = AndroidController.getNativeAccount();

            Bundle b = new Bundle();
            b.putBoolean(ContentResolver.SYNC_EXTRAS_FORCE, true);
            b.putString(SYNC_TYPE, MANUAL);
            b.putBoolean(REFRESH, true);
            b.putInt(REFRESH_DIRECTION, direction);
            if(appSource.getAuthorityType() != null) {
                b.putString(AUTHORITY_TYPE, appSource.getAuthorityType());
            }
            String authority = appSource.getAuthority();
            if (authority != null) {
                ContentResolver.requestSync(account, authority, b);
            } else {
                Log.error(TAG, "Cannot sync source with unknown authority");
            }
        }
    }

    /**
     * Requests a synchronization for multiple sources. The synchronizations
     * will be performed via the native android sync.
     */
    public void syncMultipleSources(String syncType, Vector sources) {

        Log.trace(TAG, "syncMultipleSources " + syncType);

        syncAll = true;
        refresh = false;
        syncAllType = syncType;
        syncAllSources = sources;

        AppSyncSource firstSource = null;
        if(sources.size() > 0) {
            firstSource = (AppSyncSource)syncAllSources.get(0);
            syncAllSources.removeElementAt(0);
        } else {
            Log.debug(TAG, "syncMultipleSources: no sources to synchronize");
            syncAll = false;
            return;
        }

        // We schedule only the first source here, because leaving requests in
        // the native queue will trigger the native timeout after 5 minutes.
        // We don't want this to happen, otherwise we may interrupt long syncs.
        // Successive sources will be queued in the sourceEnded method
        syncSource(syncType, firstSource);
    }
    
    /**
     * This method is redefined for Android because we create one request to the
     * native sync subsystem for each source that must be synchronized (similar
     * to what the scheduled sync does).
     */
    @Override
    public void syncAllSources(String syncType) {
        Log.info(TAG, "syncAllSources " + syncType);
        Vector sources = new Vector();
        for(int i=0;i<items.size();++i) {
            AppSyncSource appSource = (AppSyncSource)items.elementAt(i);
            if (appSource.isEnabled() && appSource.isWorking()) {
                sources.addElement(appSource);
            }
        }
        syncMultipleSources(syncType, sources);
    }

    public boolean isFirstSyncDialogDisplayed() {
        AndroidDisplayManager dm = (AndroidDisplayManager)controller.getDisplayManager();
        if (dm.isAlertPending(DisplayManager.FIRST_SYNC_DIALOG_ID)) {
            return true;
        } else {
            return false;
        }
    }

    public void sourceEnded(AppSyncSource appSource) {
        AndroidAppSyncSource next = null;
        if (syncAll) {
            // If this source was part of a sync all, then we shall trigger the next
            // sync
            if (syncAllSources.size() > 0) {
                next = (AndroidAppSyncSource)syncAllSources.elementAt(0);
                syncAllSources.removeElementAt(0);
            }
        }
        super.sourceEnded(appSource);
        if (next != null) {
            Log.info(TAG, "Requesting a new sync for: " + next.getName());
            if (refresh) {
                refresh(next.getId(), refreshDirection);
            } else {
                syncSource(syncAllType, next);
            }
        } else {
            syncAll = false;
            syncAllSources = null;
        }
    }

    /**
     * If the sync all button is pressed and there is not a syncall operation
     * we shall display a "sync in progress" message since there could be a
     * pending sync (e.g. push sync).
     */
    @Override
    public void syncAllPressed() {
        Log.trace(TAG, "Sync All Button pressed");
        String syncAllText = ((AndroidHomeScreen)homeScreen).getSyncAllText();
        if (syncAllText.equals(localization.getLanguage("home_sync_all")) && isSynchronizing()) {
            showSyncInProgressMessage();
        } else {
            super.syncAllPressed();
        }
    }

    public boolean getSyncAll() {
        return syncAll;
    }

    public void sourceFailed(AppSyncSource appSource, SyncException e) {
        syncAll = false;
        super.sourceFailed(appSource, e);
    }

    public boolean syncStarted(Vector sources) {
        if(autoSyncSwitcher.getAutoSyncMode() == AutoSyncSwitcher.AUTO_SYNC_MODE_NATIVE) {
            // Switching to custom auto sync mode
            autoSyncSwitcher.setAutoSyncMode(AutoSyncSwitcher.AUTO_SYNC_MODE_CUSTOM);
        }
        return super.syncStarted(sources);
    }

    public void syncEnded() {
        if (!syncAll) {
            super.syncEnded();
            syncAll = false;
            // Switching to custom auto sync mode
            autoSyncSwitcher.setAutoSyncMode(AutoSyncSwitcher.AUTO_SYNC_MODE_NATIVE);
        } else {
            if (syncAllSources != null && syncAllSources.size() > 0) {
                syncAll = true;
            } else {
                syncAll = false;
            }
        }
    }

    @Override
    public synchronized void continueSynchronizationAfterDialogCheck(String syncType,
                                                                     Vector filteredSources,
                                                                     boolean refresh,
                                                                     int direction,
                                                                     int delay,
                                                                     boolean fromOutside,
                                                                     boolean continueSyncFromDialog)
    {
        // If the sync shall continue after dialog check, the sync must proceed
        // through the native sync application
        if(!continueSyncFromDialog) {
            super.continueSynchronizationAfterDialogCheck(syncType, filteredSources,
                    refresh, direction, delay, fromOutside, continueSyncFromDialog);
        } else if(filteredSources.size() > 0) {
            AppSyncSource first = (AppSyncSource)filteredSources.get(0);
            if(filteredSources.size() == 1) {
                syncSingleSource(syncType, first);
            } else if(filteredSources.size() > 1) {
                this.syncAll = true;
                this.refresh = false;
                this.syncAllType = syncType;
                this.syncAllSources = filteredSources;
                syncSource(syncType, first);
            }
        }
    }

    public AndroidAppSyncSource isAnySourcePending() {

        // On Android a sync can be triggered from the native client app and
        // from within our app. When the sync has actually been fired, we are
        // sure to be always attached, but the sync may be pending. For this
        // case we need an explicit test here

        Enumeration sources = appSyncSourceManager.getWorkingSources();
        Account account = AndroidController.getNativeAccount();
        if(account == null) {
            // No account defined
            return null;
        }
        while(sources.hasMoreElements()) {
            AndroidAppSyncSource appSource = (AndroidAppSyncSource)sources.nextElement();
            String authority = appSource.getAuthority();
            if (authority != null) {
                if (ContentResolver.isSyncPending(account, authority)) {
                    return appSource;
                }
            }
        }
        return null;
    }

    public AndroidAppSyncSource isAnySourceSyncing() {

        Enumeration sources = appSyncSourceManager.getWorkingSources();
        Account account = AndroidController.getNativeAccount();
        if(account == null) {
            // No account defined
            return null;
        }
        while(sources.hasMoreElements()) {
            AndroidAppSyncSource appSource = (AndroidAppSyncSource)sources.nextElement();
            String authority = appSource.getAuthority();
            if (authority != null) {
                if (ContentResolver.isSyncActive(account, authority) &&
                   !ContentResolver.isSyncPending(account, authority))
                {
                    return appSource;
                }
            }
        }
        return null;
    }
}
