/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol". 
 */

package com.funambol.android.providers;

import java.io.FileNotFoundException;

import android.os.ParcelFileDescriptor;
import android.net.Uri;
import android.content.ContentProvider;
import android.content.Context;
import android.content.ContentValues;
import android.database.Cursor;
import android.os.Environment;
import android.provider.BaseColumns;
import android.provider.MediaStore.MediaColumns;
import android.provider.MediaStore.Images.Media;
import android.provider.MediaStore.Images;

/**
 * This is a private content provider meant to support pictures synchronization
 * in the Funambol sync client.
 * At the moment the provider is READ ONLY and it is basically a wrapper around
 * the MediaStore.
 */
public class PicturesContentProvider extends ContentProvider {

    private static final String TAG = "PicturesContentProvider";
    private static final String CONTENT_URI_STRING = "content://com.funambol.images";
    private static final String EXTERNAL_CONTENT_URI_STRING = "/external";
    private static final String INTERNAL_CONTENT_URI_STRING = "/internal";

    private static final String[] SUPPORTED_MIME_TYPES = {"image/jpeg", "image/jpg", "image/png", "image/gif"};

    public static final Uri CONTENT_URI = Uri.parse(CONTENT_URI_STRING);
    public static final Uri INTERNAL_CONTENT_URI = Uri.parse(CONTENT_URI_STRING + INTERNAL_CONTENT_URI_STRING);
    public static final Uri EXTERNAL_CONTENT_URI = Uri.parse(CONTENT_URI_STRING + EXTERNAL_CONTENT_URI_STRING);

    public static final String _ID = BaseColumns._ID;
    public static final String DISPLAY_NAME = MediaColumns.DISPLAY_NAME;
    public static final String SIZE = MediaColumns.SIZE;
    public static final String DATE_ADDED = MediaColumns.DATE_ADDED;

    private static final Uri MEDIA_PROVIDER_EXT_URI = Media.EXTERNAL_CONTENT_URI;
    private static final Uri MEDIA_PROVIDER_INT_URI = Media.INTERNAL_CONTENT_URI;

    public static final String CAMERA_IMAGE_BUCKET_NAME =
        Environment.getExternalStorageDirectory().toString() + "/DCIM/Camera";

    public static final String CAMERA_IMAGE_BUCKET_NAME_HTC =
        Environment.getExternalStorageDirectory().toString() + "/DCIM/100Media";

    public static final String CAMERA_IMAGE_BUCKET_ID     = getBucketId(CAMERA_IMAGE_BUCKET_NAME);
    public static final String CAMERA_IMAGE_BUCKET_ID_HTC = getBucketId(CAMERA_IMAGE_BUCKET_NAME_HTC);

    public PicturesContentProvider() {
    }

    public boolean onCreate() {
        return true;
    }

    public String getType(Uri uri) {
        return "vnd.android.cursor.item/vnd.funambol.images";
    }

    public int delete(Uri uri, String selection, String[] selectionArgs) {
        return 0;
    }

    public Uri insert(Uri uri, ContentValues values) {
        return uri;
    }

    public Cursor query(Uri uri, String[] projection, String selection, String[] selectionArgs, String sortOrder) {

        Context context = getContext();
        Uri mediaUri    = createMediaUri(uri);
        StringBuffer whereClause = new StringBuffer();

        // Filter pictures taken from the Camera
        whereClause.append("(");
        whereClause.append(Images.ImageColumns.BUCKET_ID).append("='" + CAMERA_IMAGE_BUCKET_ID + "'");
        whereClause.append(" OR ");
        whereClause.append(Images.ImageColumns.BUCKET_ID).append("='" + CAMERA_IMAGE_BUCKET_ID_HTC + "'");
        whereClause.append(")");

        // Filter pictures by mime type
        whereClause.append(" AND (");
        for(int i=0; i<SUPPORTED_MIME_TYPES.length; i++) {
            if(i>0)  whereClause.append(" OR ");
            whereClause.append(Images.ImageColumns.MIME_TYPE).append("='" + SUPPORTED_MIME_TYPES[i] + "'");
        }
        whereClause.append(") ");

        if (selection != null) {
            whereClause.append(" AND ").append(selection);
        }
        Cursor cursor = context.getContentResolver().query(mediaUri, projection, 
                whereClause.toString(), selectionArgs, sortOrder);
        return cursor;
    }

    public int update(Uri uri, ContentValues values, String selection, String[] selectionArgs) {
        return 0;
    }

    public ParcelFileDescriptor openFile(Uri uri, String mode) throws FileNotFoundException {
        Context context = getContext();
        Uri mediaUri    = createMediaUri(uri);
        return context.getContentResolver().openFileDescriptor(mediaUri, mode);
    }

    private Uri createMediaUri(Uri inputUri) {

        String path = inputUri.getPath();
        Uri outputUri = null;

        if (path != null) {
            if (path.startsWith(INTERNAL_CONTENT_URI_STRING)) {
                outputUri = MEDIA_PROVIDER_INT_URI;
            } else if (path.startsWith(EXTERNAL_CONTENT_URI_STRING)) {
                outputUri = MEDIA_PROVIDER_EXT_URI;
            }
        }

        if (outputUri == null) {
            throw new IllegalArgumentException("Unknown URI " + inputUri);
        }
        String lastPathSegment = inputUri.getLastPathSegment();
        if (!INTERNAL_CONTENT_URI_STRING.equals("/" + lastPathSegment) &&
            !EXTERNAL_CONTENT_URI_STRING.equals("/" + lastPathSegment))
        {
            outputUri = Uri.withAppendedPath(outputUri, lastPathSegment);
        }
        return outputUri;
    }

    public static String getBucketId(String path) {
        return String.valueOf(path.toLowerCase().hashCode());
    }
}
