/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2010 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol". 
 */

package com.funambol.android.source.media;

import android.content.Context;
import android.database.Cursor;
import android.net.Uri;
import android.provider.MediaStore.MediaColumns;
import com.funambol.android.providers.PicturesContentProvider;

import com.funambol.platform.FileAdapter;
import com.funambol.syncml.spds.SyncException;
import com.funambol.util.Log;

import java.io.IOException;


/**
 * A media metadata descritptor: combines the media resolver's iformation
 * together with the real media information on the file system. Must be
 * instantiated every time a class wants to access media file informations.
 */
public class MediaDescriptor {

    private static final String TAG_LOG = "MediaDescriptor";

    //The fields that represent the MediaDescriptor properties
    protected long added = 0;
    protected long size = 0;
    protected byte[] blobData = null;
    protected String fileName = null;

    private long fsSize  = 0;
    private long mediaStoreSize  = 0;

    Context context;
    Cursor cursor;
    Uri uri;
    String key;

    /**
     * Default constructor
     * @param context the Context object related to this instance
     * @param uri the Uri used to perform query on the cursor
     * @param key the String formatted item key
     */
    public MediaDescriptor(Context context, Uri uri, String key) {

        this.context = context;
        this.uri = Uri.withAppendedPath(uri, key);
        this.key = key;
    
        feedCursor();
    }

    private void feedCursor() {
        try {
            cursor = context.getContentResolver().query(uri, null, null, null, null);

            if (cursor == null) {
                Log.error(TAG_LOG, "Item not found " + uri);
                throw new SyncException(SyncException.CLIENT_ERROR, "Cannot get meta info for item: " + uri);
            }

            //Indexes related to the cursor's columns to be retrieved
            int addedIndex = cursor.getColumnIndexOrThrow(MediaColumns.DATE_ADDED);
            int sizeIndex = cursor.getColumnIndexOrThrow(MediaColumns.SIZE);
            int fileNameIndex = cursor.getColumnIndexOrThrow(MediaColumns.DISPLAY_NAME);
            int dataIndex = cursor.getColumnIndexOrThrow(MediaColumns.DATA);

            //Move the cursor to the first element
            cursor.moveToFirst();
            if (cursor.getCount() != 1) {
                Log.error(TAG_LOG, "Item not found, cannot compute fingerprint " + uri);
                throw new SyncException(SyncException.CLIENT_ERROR, "Cannot get meta info for item: " + uri);
            }

            added = cursor.getLong(addedIndex);

            //Retrieving the file size provided by the MediaStore could produce an
            //error on some devices. The implementation forces the usage of the File
            //System to understand the property related to the file. Here it just
            //stores the value provided by the media store for debugging purposes
            mediaStoreSize  = cursor.getLong(sizeIndex);

            //As additional control use the FileSystem to check the real size
            //of the media file.
            fileName = cursor.getString(fileNameIndex);

            //Get the byte steram content from the media store provider
            blobData = cursor.getBlob(dataIndex);

            //Read the size directly from the filesystem. Due to the fact that on
            //some devices the camera folder is differently localized, this size is
            //valid only if it is a value greater than 0.
            String fullName =
                    PicturesContentProvider.CAMERA_IMAGE_BUCKET_NAME_HTC +
                    "/" +
                    fileName;

            fsSize = getFileSize(fullName);

            if (fsSize == 0) {
                fullName =
                    PicturesContentProvider.CAMERA_IMAGE_BUCKET_NAME +
                    "/" +
                    fileName;
                fsSize = getFileSize(fullName);
            }

            size = fsSize;
            
            if (mediaStoreSize != fsSize) {
                Log.debug(TAG_LOG, "File size reported from provider is different than the one on the FS");
            }

            Log.debug(TAG_LOG, "File Name: " + fullName);
            Log.debug(TAG_LOG, "MediaStore Size: " + mediaStoreSize);
            Log.debug(TAG_LOG, "FS Size: " + fsSize);
            Log.debug(TAG_LOG, "Computed Size: " + size);
            Log.debug(TAG_LOG, "Date added: " + added);

        } catch (Exception e) {
            Log.error(TAG_LOG, "Cannot create media descriptor: ", e);
            throw new SyncException(SyncException.CLIENT_ERROR, "Cannot get meta info for item: " + uri);
        } finally {
            cursor.close();
        }
    }

    /**
     * Check the size of a file on the file system
     * @param fullName the file name with path
     * @return long the size on the file long formatted
     * @throws IOException if the FileAdapter pointing to the file encounters a
     * problem
     */
    private long getFileSize(String fullName) throws IOException {
        //Manages to find the device's file using the FS
        long pictureSize;

        FileAdapter fa = new FileAdapter(fullName);

        pictureSize = fa.getSize();

        fa = null;

        return pictureSize;
    }
}

